/*

*/

import renderer.scene.*;
import renderer.framebuffer.*;
import renderer.pipeline.*;

import java.awt.Color;

/**
   Here is a picture of a line segment, from vertex v0
   to vertex v1, that needs to be clipped four times
   before we get the final clipped line segment from
   vertex v4 to vertex v5. Remember that the clipping
   algorithm clips first on the edge {@code x = 1},
   then on {@code x = -1}, then on {@code y = 1}, and
   lastly on {@code y = -1}.
   <pre>{@code
                              v0
                              +
                          |  /
                          | /
                          |/
                          + v2
      x = -1             /|
        |               / |
        |              /  |
     ---+-------------+---+----- y = 1
        |            /v4  |
        |           /     |
        |          /      |
        |         /       |
        |        /        |
        |       /         |
        |      /          |
        |     /           |
        |    /            |
     ---+---+-------------+----- y = -1
        |  /v5            |
        | /               |
        |/                |
        + v3            x = 1
       /|
      / |
     +
    v1

   }</pre>
   <p>
   When the clipping algorithm is finished, it will have
   added four new vertices to the vertex list of this
   scene's model. And the single line segment in the model
   will only use two of the six vertices in the model's
   vertex list.
   <p>
   The clipping algorithm does not try to remove vertices
   like v0, v1, v2 and v3 which are no longer being used.
   The vertices v0 and v1 cannot be removed because they
   might still be used by other line segments in the model.
   And it is not worth the effort to try and keep track of
   vertices like v2 and v3 which are created by the clipping
   algorithm but end up not being used in the final clipped
   line segment.
   <p>
   When you run this program you can see the debugging output
   from the clipping algorithm and you can see the new vertices
   being added to the model's vertex list.
*/
public class ExtremeClippingExample
{
   public static void main(String[] args)
   {
      final Scene scene = new Scene();
      final Model model = new Model("extreme clipping example");
      scene.addModel(model);

      model.addVertex(new Vertex( 1.25,  1.5, -1),
                      new Vertex(-1.25, -1.5, -1));

      model.addLineSegment(new LineSegment(0, 1));

      Pipeline.debug = true;
      Clip.debug = true;
      System.out.println(model);

      final FrameBuffer fb = new FrameBuffer(100, 100);
      fb.clearFB(Color.darkGray);
      Pipeline.render(scene, fb);
      fb.dumpFB2File("ExtremeClippingExample.ppm");
   }
}

/*
Exercise: Suppose that the order of the edges used by the
clipping algorithm is
   x =  1,
   y = -1,
   x = -1,
   y =  1.
How many times would the line in the above example be clipped?

Using this second order for the edges in the clipping algorithm,
find an example of a line that will be clipped four times.
*/
