/*

*/

package renderer.pipeline;
import  renderer.scene.*;

import java.util.List;
import java.util.ArrayList;
import java.util.Optional;

/**
   Clip a (projected) {@link LineSegment} that sticks out of the
   view rectangle in the image plane.
<p>
   This clipping algorithm is a simplification of the Liang-Barsky
   Parametric Line Clipping algorithm.
<p>
   This algorithm assumes that all {@code Vertex} objects have been
   projected onto the {@link renderer.scene.Camera}'s image plane,
   {@code z = -1}. This algorithm also assumes that the camera's view
   rectangle in the image plane is
   <pre>{@code
      -1 <= x <= +1  and
      -1 <= y <= +1.
   }</pre>
<p>
   If a line segment's projected vertex has an {@code x} or {@code y}
   coordinate with absolute value greater than 1, then that vertex
   "sticks out" of the view rectangle. This algorithm will clip the
   line segment so that both of the line segment's vertices are within
   the view rectangle.
<p>
   Here is an outline of the clipping algorithm.
<p>
   Process each line segment, using the following steps.
<p>
     1) Test if the line segment no longer needs to be clipped, i.e.,
        both of its vertices are within the view rectangle. If this is
        the case, then return the line segment wrapped in an
        {@link Optional} object.
        <pre>{@code
               x=-1        x=+1
                 |          |
                 |          |
             ----+----------+----- y = +1
                 |     v1   |
                 |    /     |
                 |   /      |
                 |  /       |
                 | v0       |
             ----+----------+----- y = -1
                 |          |
                 |          |
        }</pre>
<p>
     2) Test if the line segment should be "trivially rejected". A line
        segment is "trivially rejected" if it is on the wrong side of any
        of the four lines that bound the view rectangle (i.e., the four
        lines {@code x = 1}, {@code x = -1}, {@code y = 1}, {@code y = -1}).
        If so, then return an empty {@link Optional} object indicating that
        the line segment is not be rasterized into the framebuffer.
<p>
        Notice that a line like the following one is trivially rejected
        because it is on the "wrong" side of the line {@code x = 1}.
        <pre>{@code
                           x=1
                            |            v1
                            |            /
                 +----------+           /
                 |          |          /
                 |          |         /
                 |          |        /
                 |          |       /
                 |          |      /
                 +----------+     /
                            |    /
                            |  v0
        }</pre>
        But the following line is NOT trivially rejected because, even
        though it is completely outside of the view rectangle, this line
        is not entirely on the wrong side of any one of the four lines
        {@code x = 1}, {@code x = -1}, {@code y = 1}, or {@code y = -1}.
        The line below will get clipped at least one time (either on the
        line {@code x = 1} or the line {@code y = -1}) before it is
        (recursively) a candidate for "trivial rejection". Notice that
        the line below could even be clipped twice, first on {@code y = 1},
        then on {@code x = 1}, before it can be trivially rejected (by
        being on the wrong side of {@code y = -1}).
        <pre>{@code
                           x=1
                            |          v1
                            |         /
                 +----------+        /
                 |          |       /
                 |          |      /
                 |          |     /
                 |          |    /
                 |          |   /
                 +----------+  /
                            | /
                            |/
                            /
                           /|
                          / |
                        v0
        }</pre>
<p>
     3) If the line segment has been neither accepted nor rejected, then
        it needs to be clipped. So we succesively test the line segment
        against each of the four clipping lines, {@code x = 1}, {@code x = -1},
        {@code y = 1}, and {@code y = -1}, to determine if the line segment
        crosses one of those lines. We clip the line segment against each
        line which we find that it crosses. After each clip, the line is
        checked to see if it can be trivally accepted or trivally rejected.
<p>
        When we clip a line segment against a clipping line, it is always
        the case that one endpoint of the line segment is on the "right"
        side of the clipping line and the other endpoint is on the "wrong"
        side of the clipping line. In the following picture, assume that
        {@code v0} is on the "wrong" side of the clipping line({@code x = 1})
        and {@code v1} is on the "right" side. So {@code v0} needs to be
        clipped off the line segment and replaced by a new vertex.
        <pre>{@code
                             x=1
                         v1   |
                           \  |
                            \ |
                             \|
                              \
                              |\
                              | \
                              |  \
                              |   v0
        }</pre>
        Represent points {@code p(t)} on the line segment between {@code v0}
        and {@code v1} with the following parametric equation.
        <pre>{@code
                  p(t) = (1-t) * v0 + t * v1  with  0 <= t <= 1
        }</pre>
        Notice that this equation parameterizes the line segment starting
        with {@code v0} at {@code t=0} (on the "wrong side") and ending
        with {@code v1} at {@code t=1} (on the "right side"). We need to
        find the value of {@code t} when the line segment crosses the
        clipping line {@code x = 1}. Let {@code v0 = (x0, y0)} and let
        {@code v1 = (x1, y1)}. Then the above parametric equation becomes
        the two component equations
        <pre>{@code
                 x(t) = (1-t) * x0 + t * x1,
                 y(t) = (1-t) * y0 + t * y1,  with  0 <= t <= 1.
        }</pre>
        Since the clipping line in this example is {@code x = 1}, we need
        to solve the equation {@code x(t) = 1} for {@code t}. So we need
        to solve
        <pre>{@code
                  1 = (1-t) * x0 + t * x1
        }</pre>
        for {@code t}. Here are a few algebra steps.
        <pre>{@code
                  1 = x0 - t * x0 + t * x1
                  1 = x0 + t * (x1 - x0)
                  1 - x0 = t * (x1 - x0)
                       t = (1 - x0)/(x1 - x0)
        }</pre>
        We get similar equations for {@code t} if we clip against the other
        clipping lines ({@code x = -1}, {@code y = 1}, or {@code y = -1})
        and we assume that {@code v0} is on the "wrong side" and {@code v1}
        is on the "right side".
<p>
        Let {@code t0} denote the above value for {@code t}. With this value
        for {@code t}, we can compute the y-coordinate of the new vertex
        {@code p(t0)} that replaces {@code v0}.
        <pre>{@code
                             x=1
                        v1    |
                          \   |
                           \  |
                            \ |
                              p(t0)=(1, y(t0))
                              |
                              |
                              |
         }</pre>
         Here is the algebra.
         <pre>{@code
                  y(t0) = (1-t0) * y0 + t0 * y1
                        = y0 + t0 * (y1 - y0)
                        = y0 + (1 - x0)*((y1 - y0)/(x1 - x0))
         }</pre>
         Finally, the new line segment between {@code v1} and the new
         vertex {@code p(t0)} is checked to see if it should be trivially
         accepted or trivially rejected. If it is neither accepted nor
         rejected, we go on to the next clipping equation.
*/
public final class Clip_v4
{
   public static boolean debug = false;

   /**
      Start with a {@link Model} that contains {@link LineSegment}s
      that have been projected onto the camera's view plane, {@code z = -1}.
   <p>
      If a projected {@link LineSegment} sticks out of the camera's
      view rectangle, then replace that {@link LineSegment}, in the
      {@link Model}'s list of line segments, with one that has been
      clipped so that it is contained in the view rectangle.
   <p>
      If a projected {@link LineSegment} is completely outside of
      the view rectangle, then drop that {@link LineSegment} from
      the {@link Model}'s list of line segments.
   <p>
      Return a {@link Model} for which every {@link LineSegment} is
      completely contained in the camera's view rectangle.

      @param model  {@link Model} containing projected {@link LineSegment}s
      @return a {@link Model} containing {@link LineSegment}s clipped to the view rectangle
   */
   public static Model clip(final Model model)
   {
      boolean debug = Clip.debug && (Pipeline.debug || model.debug);

      final List<LineSegment> newLineSegmentList = new ArrayList<>();

      for (final LineSegment ls : model.lineSegmentList)
      {
         Pipeline.logLineSegment("2. Clipping", model, ls);

         Optional<LineSegment> ls_clipped = clip(model, ls);

         if ( ls_clipped.isPresent() )
         {
            // Keep the line segments that are visible.
            newLineSegmentList.add( ls_clipped.get() );

            Pipeline.logLineSegment("2. Clipping (accept)", model, ls_clipped.get());
         }
         else
         {
            // Discard the line segments that are not visible.
            Pipeline.logLineSegment("2. Clipping (reject)", model, ls);
         }
      }
      // Replace the model's original list of line segments
      // with the list of clipped line segments.
      return new Model(model.vertexList, // has been updated with clipped vertices
                       newLineSegmentList,
                       model.name,
                       model.visible,
                       model.debug);
   }


   /**
      If the {@link LineSegment} sticks out of the view rectangle,
      then return a clipped version that is contained in the view
      rectangle. The new, clipped {@link LineSegment} object is
      returned wrapped in an {@link Optional} object.
      <p>
      At least one new clipped {@link Vertex} will be added to the
      {@link Model}'s vertex list (and as many as four new vertices
      may be added to the {@link Model}'s vertex list).
      <p>
      If the {@link LineSegment} is completely outside of the view
      rectangle, then return an empty {@link Optional} object to
      indicate that the {@link LineSegment} should be discarded.

      @param model  {@link Model} that the {@link LineSegment} {@code ls} comes from
      @param ls     {@link LineSegment} to be clipped
      @return a clipped version of {@code ls} wrapped in an {@link Optional} object
   */
   private static Optional<LineSegment> clip(final Model model, final LineSegment ls)
   {
      boolean debug = Clip.debug && (Pipeline.debug || model.debug);

      // Make local copies of several values.
      final Vertex v0 = model.vertexList.get( ls.index[0] );
      final Vertex v1 = model.vertexList.get( ls.index[1] );

      final double x0 = v0.x,  y0 = v0.y;
      final double x1 = v1.x,  y1 = v1.y;

      // 1. Check for trivial accept.
      if ( ! ( Math.abs( x0 ) > 1
            || Math.abs( y0 ) > 1
            || Math.abs( x1 ) > 1
            || Math.abs( y1 ) > 1 ) )
      {
         if (debug) System.err.println("-Trivial accept.");
         return Optional.of(ls); // better than "return ls;"
      }
      // 2. Check for trivial delete.
      if ( (x0 >  1 && x1 >  1)   // to the right of the line x = 1
        || (x0 < -1 && x1 < -1)   // to the left of the line x = -1
        || (y0 >  1 && y1 >  1)   // above the line y = 1
        || (y0 < -1 && y1 < -1) ) // below the line y = -1
      {
         if (debug) System.err.println("-Trivial delete.");
         return Optional.empty(); // better than "return null;"
      }
      // 3. Need to clip this line segment.
      if (x0 > 1)  // ls crosses the line x = 1
      {
         final String equation = "x = +1";
         final String vOutside = "v0";
         final double vOx = x0,  vOy = y0;
         final double vIx = x1,  vIy = y1;
         final double t = (1 - vOx) / (vIx - vOx);
         final double x = 1;
         final double y = (1-t) * vOy + t * vIy;
         final Vertex newVertex = new Vertex(x, y, 0);
         // Modify the Model to contain the new Vertex.
         final int newIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
         // Create a new LineSegment using the new Vertex
         // and keeping the old LineSegment's inside Vertex.
         final LineSegment newLS = new LineSegment(newIndex, ls.index[1]);

         if (debug)
         {
            System.err.printf("-Clip off %s at %s\n", vOutside, equation);
            System.err.printf("- t = % .25f\n", t);
            System.err.printf("- <x_i,y_i> = <% .24f % .24f\n",  vIx, vIy);
            System.err.printf("- <x_o,y_o> = <% .24f % .24f\n",  vOx, vOy);
            System.err.printf("- <x_c,y_c> = <% .24f % .24f\n",    x,   y);
         }

         // Check for trivial accept.
         if ( ! ( Math.abs( x  ) > 1
               || Math.abs( y  ) > 1
               || Math.abs( x1 ) > 1
               || Math.abs( y1 ) > 1 ) )
         {
            if (debug) System.err.println("-Trivial accept.");
            return Optional.of(newLS); // better than "return ls;"
         }
         // Check for trivial delete.
         else if ( (x  >  1 && x1 >  1)   // to the right of the line x = 1
                || (x  < -1 && x1 < -1)   // to the left of the line x = -1
                || (y  >  1 && y1 >  1)   // above the line y = 1
                || (y  < -1 && y1 < -1) ) // below the line y = -1
         {
            if (debug) System.err.println("-Trivial delete.");
            return Optional.empty(); // better than "return null;"
         }
      }

      if (x1 > 1)  // ls crosses the line x = 1
      {
         final String equation = "x = +1";
         final String vOutside = "v1";
         final double vIx = x0,  vIy = y0;
         final double vOx = x1,  vOy = y1;
         final double t = (1 - vOx) / (vIx - vOx);
         final double x = 1;  // prevent rounding errors
         final double y = (1-t) * vOy + t * vIy;
         final Vertex newVertex = new Vertex(x, y, 0);
         final int newIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
         final LineSegment newLS = new LineSegment(ls.index[0], newIndex);

         if (debug)
         {
            System.err.printf("-Clip off %s at %s\n", vOutside, equation);
            System.err.printf("- t = % .25f\n", t);
            System.err.printf("- <x_i,y_i> = <% .24f % .24f\n",  vIx, vIy);
            System.err.printf("- <x_o,y_o> = <% .24f % .24f\n",  vOx, vOy);
            System.err.printf("- <x_c,y_c> = <% .24f % .24f\n",    x,   y);
         }

         // Check for trivial accept.
         if ( ! ( Math.abs( x0 ) > 1
               || Math.abs( y0 ) > 1
               || Math.abs( x  ) > 1
               || Math.abs( y  ) > 1 ) )
         {
            if (debug) System.err.println("-Trivial accept.");
            return Optional.of(newLS);
         }
         // Check for trivial delete.
         else if ( (x0 >  1 && x  >  1)   // to the right of the line x = 1
                || (x0 < -1 && x  < -1)   // to the left of the line x = -1
                || (y0 >  1 && y  >  1)   // above the line y = 1
                || (y0 < -1 && y  < -1) ) // below the line y = -1
         {
            if (debug) System.err.println("-Trivial delete.");
            return Optional.empty();
         }
      }

      if (x0 < -1)  // ls crosses the line x = -1
      {
         final String equation = "x = -1";
         final String vOutside = "v0";
         final double vOx = x0,  vOy = y0;
         final double vIx = x1,  vIy = y1;
         final double t = (-1 - vOx) / (vIx - vOx);
         final double x = -1;  // prevent rounding errors
         final double y = (1-t) * vOy + t * vIy;
         final Vertex newVertex = new Vertex(x, y, 0);
         final int newIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
         final LineSegment newLS = new LineSegment(newIndex, ls.index[1]);

         if (debug)
         {
            System.err.printf("-Clip off %s at %s\n", vOutside, equation);
            System.err.printf("- t = % .25f\n", t);
            System.err.printf("- <x_i,y_i> = <% .24f % .24f\n",  vIx, vIy);
            System.err.printf("- <x_o,y_o> = <% .24f % .24f\n",  vOx, vOy);
            System.err.printf("- <x_c,y_c> = <% .24f % .24f\n",    x,   y);
         }

         // Check for trivial accept.
         if ( ! ( Math.abs( x  ) > 1
               || Math.abs( y  ) > 1
               || Math.abs( x1 ) > 1
               || Math.abs( y1 ) > 1 ) )
         {
            if (debug) System.err.println("-Trivial accept.");
            return Optional.of(newLS);
         }
         // Check for trivial delete.
         else if ( (x  >  1 && x1 >  1)   // to the right of the line x = 1
                || (x  < -1 && x1 < -1)   // to the left of the line x = -1
                || (y  >  1 && y1 >  1)   // above the line y = 1
                || (y  < -1 && y1 < -1) ) // below the line y = -1
         {
            if (debug) System.err.println("-Trivial delete.");
            return Optional.empty();
         }
      }

      if (x1 < -1)  // ls crosses the line x = -1
      {
         final String equation = "x = -1";
         final String vOutside = "v1";
         final double vIx = x0,  vIy = y0;
         final double vOx = x1,  vOy = y1;
         final double t = (-1 - vOx) / (vIx - vOx);
         final double x = -1;  // prevent rounding errors
         final double y = (1-t) * vOy + t * vIy;
         final Vertex newVertex = new Vertex(x, y, 0);
         final int newIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
         final LineSegment newLS = new LineSegment(ls.index[0], newIndex);

         if (debug)
         {
            System.err.printf("-Clip off %s at %s\n", vOutside, equation);
            System.err.printf("- t = % .25f\n", t);
            System.err.printf("- <x_i,y_i> = <% .24f % .24f\n",  vIx, vIy);
            System.err.printf("- <x_o,y_o> = <% .24f % .24f\n",  vOx, vOy);
            System.err.printf("- <x_c,y_c> = <% .24f % .24f\n",    x,   y);
         }

         // Check for trivial accept.
         if ( ! ( Math.abs( x0 ) > 1
               || Math.abs( y0 ) > 1
               || Math.abs( x  ) > 1
               || Math.abs( y  ) > 1 ) )
         {
            if (debug) System.err.println("-Trivial accept.");
            return Optional.of(newLS);
         }
         // Check for trivial delete.
         else if ( (x0 >  1 && x  >  1)   // to the right of the line x = 1
                || (x0 < -1 && x  < -1)   // to the left of the line x = -1
                || (y0 >  1 && y  >  1)   // above the line y = 1
                || (y0 < -1 && y  < -1) ) // below the line y = -1
         {
            if (debug) System.err.println("-Trivial delete.");
            return Optional.empty();
         }
      }

      if (y0 > 1)  // ls crosses the line y = 1
      {
         final String equation = "y = +1";
         final String vOutside = "v0";
         final double vOx = x0,  vOy = y0;
         final double vIx = x1,  vIy = y1;
         final double t = (1 - vOy) / (vIy - vOy);
         final double x = (1-t) * vOx + t * vIx;
         final double y = 1;  // prevent rounding errors
         final Vertex newVertex = new Vertex(x, y, 0);
         final int newIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
         final LineSegment newLS = new LineSegment(newIndex, ls.index[1]);

         if (debug)
         {
            System.err.printf("-Clip off %s at %s\n", vOutside, equation);
            System.err.printf("- t = % .25f\n", t);
            System.err.printf("- <x_i,y_i> = <% .24f % .24f\n",  vIx, vIy);
            System.err.printf("- <x_o,y_o> = <% .24f % .24f\n",  vOx, vOy);
            System.err.printf("- <x_c,y_c> = <% .24f % .24f\n",    x,   y);
         }

         // Check for trivial accept.
         if ( ! ( Math.abs( x  ) > 1
               || Math.abs( y  ) > 1
               || Math.abs( x1 ) > 1
               || Math.abs( y1 ) > 1 ) )
         {
            if (debug) System.err.println("-Trivial accept.");
            return Optional.of(newLS);
         }
         // Check for trivial delete.
         else if ( (x  >  1 && x1 >  1)   // to the right of the line x = 1
                || (x  < -1 && x1 < -1)   // to the left of the line x = -1
                || (y  >  1 && y1 >  1)   // above the line y = 1
                || (y  < -1 && y1 < -1) ) // below the line y = -1
         {
            if (debug) System.err.println("-Trivial delete.");
            return Optional.empty();
         }
      }

      if (y1 > 1)  // ls crosses the line y = 1
      {
         final String equation = "y = +1";
         final String vOutside = "v1";
         final double vIx = x0,  vIy = y0;
         final double vOx = x1,  vOy = y1;
         final double t = (1 - vOy) / (vIy - vOy);
         final double x = (1-t) * vOx + t * vIx;
         final double y = 1;  // prevent rounding errors
         final Vertex newVertex = new Vertex(x, y, 0);
         final int newIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
         final LineSegment newLS = new LineSegment(ls.index[0], newIndex);

         if (debug)
         {
            System.err.printf("-Clip off %s at %s\n", vOutside, equation);
            System.err.printf("- t = % .25f\n", t);
            System.err.printf("- <x_i,y_i> = <% .24f % .24f\n",  vIx, vIy);
            System.err.printf("- <x_o,y_o> = <% .24f % .24f\n",  vOx, vOy);
            System.err.printf("- <x_c,y_c> = <% .24f % .24f\n",    x,   y);
         }

         // Check for trivial accept.
         if ( ! ( Math.abs( x0 ) > 1
               || Math.abs( y0 ) > 1
               || Math.abs( x  ) > 1
               || Math.abs( y  ) > 1 ) )
         {
            if (debug) System.err.println("-Trivial accept.");
            return Optional.of(newLS);
         }
         // Check for trivial delete.
         else if ( (x0 >  1 && x  >  1)   // to the right of the line x = 1
                || (x0 < -1 && x  < -1)   // to the left of the line x = -1
                || (y0 >  1 && y  >  1)   // above the line y = 1
                || (y0 < -1 && y  < -1) ) // below the line y = -1
         {
            if (debug) System.err.println("-Trivial delete.");
            return Optional.empty();
         }
      }

      if (y0 < -1)  // ls crosses the line y = -1
      {
         final String equation = "y = -1";
         final String vOutside = "v0";
         final double vOx = x0,  vOy = y0;
         final double vIx = x1,  vIy = y1;
         final double t = (-1 - vOy) / (vIy - vOy);
         final double x = (1-t) * vOx + t * vIx;
         final double y = -1;  // prevent rounding errors
         final Vertex newVertex = new Vertex(x, y, 0);
         final int newIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
         final LineSegment newLS = new LineSegment(newIndex, ls.index[1]);

         if (debug)
         {
            System.err.printf("-Clip off %s at %s\n", vOutside, equation);
            System.err.printf("- t = % .25f\n", t);
            System.err.printf("- <x_i,y_i> = <% .24f % .24f\n",  vIx, vIy);
            System.err.printf("- <x_o,y_o> = <% .24f % .24f\n",  vOx, vOy);
            System.err.printf("- <x_c,y_c> = <% .24f % .24f\n",    x,   y);
         }

         // Check for trivial accept.
         if ( ! ( Math.abs( x  ) > 1
               || Math.abs( y  ) > 1
               || Math.abs( x1 ) > 1
               || Math.abs( y1 ) > 1 ) )
         {
            if (debug) System.err.println("-Trivial accept.");
            return Optional.of(newLS);
         }
         // Check for trivial delete.
         else if ( (x  >  1 && x1 >  1)   // to the right of the line x = 1
                || (x  < -1 && x1 < -1)   // to the left of the line x = -1
                || (y  >  1 && y1 >  1)   // above the line y = 1
                || (y  < -1 && y1 < -1) ) // below the line y = -1
         {
            if (debug) System.err.println("-Trivial delete.");
            return Optional.empty();
         }
      }

      if (y1 < -1)  // ls crosses the line y = -1
      {
         final String equation = "y = -1";
         final String vOutside = "v1";
         final double vIx = x0,  vIy = y0;
         final double vOx = x1,  vOy = y1;
         final double t = (-1 - vOy) / (vIy - vOy);
         final double x = (1-t) * vOx + t * vIx;
         final double y = -1;  // prevent rounding errors
         final Vertex newVertex = new Vertex(x, y, 0);
         final int newIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
         final LineSegment newLS = new LineSegment(ls.index[0], newIndex);

         if (debug)
         {
            System.err.printf("-Clip off %s at %s\n", vOutside, equation);
            System.err.printf("- t = % .25f\n", t);
            System.err.printf("- <x_i,y_i> = <% .24f % .24f\n",  vIx, vIy);
            System.err.printf("- <x_o,y_o> = <% .24f % .24f\n",  vOx, vOy);
            System.err.printf("- <x_c,y_c> = <% .24f % .24f\n",    x,   y);
         }

         // Check for trivial accept.
         if ( ! ( Math.abs( x0 ) > 1
               || Math.abs( y0 ) > 1
               || Math.abs( x  ) > 1
               || Math.abs( y  ) > 1 ) )
         {
            if (debug) System.err.println("-Trivial accept.");
            return Optional.of(newLS);
         }
         // Check for trivial delete.
         else if ( (x0 >  1 && x  >  1)   // to the right of the line x = 1
                || (x0 < -1 && x  < -1)   // to the left of the line x = -1
                || (y0 >  1 && y  >  1)   // above the line y = 1
                || (y0 < -1 && y  < -1) ) // below the line y = -1
         {
            if (debug) System.err.println("-Trivial delete.");
            return Optional.empty();
         }
      }

      // Should never get here!
      return Optional.empty();
   }
}
